#
# VecHelp.py -- help classes for vector drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

IMAGE = 0
LINE = 1
CIRCLE = 2
BEZIER = 3
ELLIPSE_BEZIER = 4
POLYGON = 5
PATH = 6
TEXT = 7
